import unittest

from kafka.tools.protocol.requests.topic_metadata_v4 import TopicMetadataV4Request


class TopicMetadataV1RequestTest(unittest.TestCase):
    def test_process_arguments(self):
        val = TopicMetadataV4Request.process_arguments(['topicname', 'anothertopic'])
        assert val == {'topics': ['topicname', 'anothertopic'], 'allow_auto_topic_creation': False}

    def test_process_arguments_all(self):
        val = TopicMetadataV4Request.process_arguments([])
        assert val == {'topics': None, 'allow_auto_topic_creation': False}



import abc
import pprint
import six


def _decode_plain_type(value_type, buf):
    if value_type == 'int8':
        return buf.getInt8()
    elif value_type == 'int16':
        return buf.getInt16()
    elif value_type == 'int32':
        return buf.getInt32()
    elif value_type == 'int64':
        return buf.getInt64()
    elif value_type == 'string':
        val_len = buf.getInt16()
        return None if val_len == -1 else buf.get(val_len).decode("utf-8")
    elif value_type == 'bytes':
        val_len = buf.getInt32()
        return None if val_len == -1 else buf.get(val_len)
    elif value_type == 'boolean':
        return buf.getInt8() == 1
    else:
        raise NotImplementedError("Reference to non-implemented type in schema: {0}".format(value_type))


def _decode_array(array_schema, buf):
    array_len = buf.getInt32()
    if array_len == -1:
        return None

    if isinstance(array_schema, six.string_types):
        return [_decode_plain_type(array_schema, buf) for i in range(array_len)]
    else:
        return [_decode_sequence(array_schema, buf) for i in range(array_len)]


def _decode_sequence(sequence_schema, buf):
    val = {}
    for entry in sequence_schema:
        if entry['type'].lower() == 'array':
            val[entry['name']] = _decode_array(entry['item_type'], buf)
        else:
            val[entry['name']] = _decode_plain_type(entry['type'].lower(), buf)
    return val


@six.add_metaclass(abc.ABCMeta)
class BaseResponse():  # pragma: no cover
    @abc.abstractproperty
    def schema(self):
        pass

    @classmethod
    def from_bytebuffer(cls, correlation_id, buf):
        seq_obj = _decode_sequence(cls.schema, buf)
        rv = cls(seq_obj)
        rv.correlation_id = correlation_id
        return rv

    def __init__(self, sequence_obj):
        self._response = sequence_obj

    def __hash__(self):
        return id(self)

    def __str__(self):
        pp = pprint.PrettyPrinter(indent=4)
        return pp.pformat(self._response)

    def __len__(self):
        return len(self._response)

    def __contains__(self, k):
        return k in self._response

    def __getitem__(self, k):
        return self._response[k]

    def __setitem__(self, k):
        raise NotImplementedError

    def __delitem__(self, k):
        raise NotImplementedError



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.list_offset_v1 import ListOffsetV1Request, _parse_next_topic
from kafka.tools.protocol.responses.list_offset_v2 import ListOffsetV2Response


class ListOffsetV2Request(ListOffsetV1Request):
    api_version = 2
    response = ListOffsetV2Response
    cmd = "ListOffset"

    help_string = ("Request:     {0}V{1}\n".format(cmd, api_version) +
                   "Format:      {0}V{1} replica_id committed (topic (partition,timestamp ...) ...)\n".format(cmd, api_version) +
                   "             replica_id should be -1 for normal consumers\n" +
                   "Description: Send replica information to broker\n" +
                   "Examples:    {0}V{1} -1 ExampleTopic 0,1477613177000 1,-1 ExampleTopic2 0,-2\n".format(cmd, api_version))

    schema = [
        {'name': 'replica_id', 'type': 'int32'},
        {'name': 'topics',
         'type': 'array',
         'item_type': [
             {'name': 'topic', 'type': 'string'},
             {'name': 'partitions',
              'type': 'array',
              'item_type': [
                  {'name': 'partition', 'type': 'int32'},
                  {'name': 'timestamp', 'type': 'int64'},
              ]},
         ]},
    ]

    @classmethod
    def process_arguments(cls, cmd_args):
        if len(cmd_args) < 4:
            raise ArgumentError("ListOffsetV2 requires at least 4 arguments")
        if cmd_args[1] not in ('true', 'false'):
            raise ArgumentError("committed must be 'true' or 'false'")

        try:
            replica_id = int(cmd_args.pop(0))
        except ValueError:
            raise ArgumentError("The replica_id must be an integer")
        committed = cmd_args.pop(0)

        values = {'replica_id': replica_id,
                  'isolation_level': 1 if committed == 'true' else 0,
                  'topics': []}

        while len(cmd_args) > 0:
            topic, cmd_args = _parse_next_topic(cmd_args)
            values['topics'].append(topic)

        return values



import unittest
from tests.tools.protocol.utilities import validate_schema

from kafka.tools.protocol.responses.alter_configs_v0 import AlterConfigsV0Response
from kafka.tools.protocol.responses.alter_replica_log_dirs_v0 import AlterReplicaLogDirsV0Response
from kafka.tools.protocol.responses.api_versions_v0 import ApiVersionsV0Response
from kafka.tools.protocol.responses.api_versions_v1 import ApiVersionsV1Response
from kafka.tools.protocol.responses.controlled_shutdown_v0 import ControlledShutdownV0Response
from kafka.tools.protocol.responses.controlled_shutdown_v1 import ControlledShutdownV1Response
from kafka.tools.protocol.responses.create_acls_v0 import CreateAclsV0Response
from kafka.tools.protocol.responses.create_partitions_v0 import CreatePartitionsV0Response
from kafka.tools.protocol.responses.create_topics_v0 import CreateTopicsV0Response
from kafka.tools.protocol.responses.create_topics_v1 import CreateTopicsV1Response
from kafka.tools.protocol.responses.create_topics_v2 import CreateTopicsV2Response
from kafka.tools.protocol.responses.delete_acls_v0 import DeleteAclsV0Response
from kafka.tools.protocol.responses.delete_topics_v0 import DeleteTopicsV0Response
from kafka.tools.protocol.responses.delete_topics_v1 import DeleteTopicsV1Response
from kafka.tools.protocol.responses.delete_records_v0 import DeleteRecordsV0Response
from kafka.tools.protocol.responses.describe_acls_v0 import DescribeAclsV0Response
from kafka.tools.protocol.responses.describe_configs_v0 import DescribeConfigsV0Response
from kafka.tools.protocol.responses.describe_groups_v0 import DescribeGroupsV0Response
from kafka.tools.protocol.responses.describe_groups_v1 import DescribeGroupsV1Response
from kafka.tools.protocol.responses.describe_log_dirs_v0 import DescribeLogDirsV0Response
from kafka.tools.protocol.responses.fetch_v0 import FetchV0Response
from kafka.tools.protocol.responses.fetch_v1 import FetchV1Response
from kafka.tools.protocol.responses.fetch_v2 import FetchV2Response
from kafka.tools.protocol.responses.group_coordinator_v0 import GroupCoordinatorV0Response
from kafka.tools.protocol.responses.find_coordinator_v1 import FindCoordinatorV1Response
from kafka.tools.protocol.responses.heartbeat_v0 import HeartbeatV0Response
from kafka.tools.protocol.responses.heartbeat_v1 import HeartbeatV1Response
from kafka.tools.protocol.responses.join_group_v0 import JoinGroupV0Response
from kafka.tools.protocol.responses.join_group_v1 import JoinGroupV1Response
from kafka.tools.protocol.responses.join_group_v2 import JoinGroupV2Response
from kafka.tools.protocol.responses.leader_and_isr_v0 import LeaderAndIsrV0Response
from kafka.tools.protocol.responses.leader_and_isr_v1 import LeaderAndIsrV1Response
from kafka.tools.protocol.responses.leave_group_v0 import LeaveGroupV0Response
from kafka.tools.protocol.responses.leave_group_v1 import LeaveGroupV1Response
from kafka.tools.protocol.responses.list_groups_v0 import ListGroupsV0Response
from kafka.tools.protocol.responses.list_groups_v1 import ListGroupsV1Response
from kafka.tools.protocol.responses.list_offset_v0 import ListOffsetV0Response
from kafka.tools.protocol.responses.list_offset_v1 import ListOffsetV1Response
from kafka.tools.protocol.responses.list_offset_v2 import ListOffsetV2Response
from kafka.tools.protocol.responses.member_assignment_v0 import MemberAssignmentV0
from kafka.tools.protocol.responses.metadata_v0 import MetadataV0Response
from kafka.tools.protocol.responses.metadata_v1 import MetadataV1Response
from kafka.tools.protocol.responses.metadata_v2 import MetadataV2Response
from kafka.tools.protocol.responses.metadata_v3 import MetadataV3Response
from kafka.tools.protocol.responses.metadata_v4 import MetadataV4Response
from kafka.tools.protocol.responses.metadata_v5 import MetadataV5Response
from kafka.tools.protocol.responses.offset_commit_v0 import OffsetCommitV0Response
from kafka.tools.protocol.responses.offset_commit_v1 import OffsetCommitV1Response
from kafka.tools.protocol.responses.offset_commit_v2 import OffsetCommitV2Response
from kafka.tools.protocol.responses.offset_commit_v3 import OffsetCommitV3Response
from kafka.tools.protocol.responses.offset_fetch_v0 import OffsetFetchV0Response
from kafka.tools.protocol.responses.offset_fetch_v1 import OffsetFetchV1Response
from kafka.tools.protocol.responses.offset_fetch_v2 import OffsetFetchV2Response
from kafka.tools.protocol.responses.offset_fetch_v3 import OffsetFetchV3Response
from kafka.tools.protocol.responses.offsets_for_leader_epoch_v0 import OffsetForLeaderEpochV0Response
from kafka.tools.protocol.responses.produce_v0 import ProduceV0Response
from kafka.tools.protocol.responses.produce_v1 import ProduceV1Response
from kafka.tools.protocol.responses.produce_v2 import ProduceV2Response
from kafka.tools.protocol.responses.sasl_authenticate_v0 import SaslAuthenticateV0Response
from kafka.tools.protocol.responses.sasl_handshake_v0 import SaslHandshakeV0Response
from kafka.tools.protocol.responses.sasl_handshake_v1 import SaslHandshakeV1Response
from kafka.tools.protocol.responses.stop_replica_v0 import StopReplicaV0Response
from kafka.tools.protocol.responses.sync_group_v0 import SyncGroupV0Response
from kafka.tools.protocol.responses.sync_group_v1 import SyncGroupV1Response
from kafka.tools.protocol.responses.update_metadata_v0 import UpdateMetadataV0Response
from kafka.tools.protocol.responses.update_metadata_v1 import UpdateMetadataV1Response
from kafka.tools.protocol.responses.update_metadata_v2 import UpdateMetadataV2Response
from kafka.tools.protocol.responses.update_metadata_v3 import UpdateMetadataV3Response
from kafka.tools.protocol.responses.update_metadata_v4 import UpdateMetadataV4Response


class ValidateResponseSchemaTests(unittest.TestCase):
    def test_schema_validation(self):
        for klass in (
                AlterConfigsV0Response,
                AlterReplicaLogDirsV0Response,
                ApiVersionsV0Response,
                ApiVersionsV1Response,
                ControlledShutdownV0Response,
                ControlledShutdownV1Response,
                CreateAclsV0Response,
                CreatePartitionsV0Response,
                CreateTopicsV0Response,
                CreateTopicsV1Response,
                CreateTopicsV2Response,
                DeleteAclsV0Response,
                DeleteTopicsV0Response,
                DeleteTopicsV1Response,
                DeleteRecordsV0Response,
                DescribeAclsV0Response,
                DescribeConfigsV0Response,
                DescribeGroupsV0Response,
                DescribeGroupsV1Response,
                DescribeLogDirsV0Response,
                FetchV0Response,
                FetchV1Response,
                FetchV2Response,
                GroupCoordinatorV0Response,
                FindCoordinatorV1Response,
                HeartbeatV0Response,
                HeartbeatV1Response,
                JoinGroupV0Response,
                JoinGroupV1Response,
                JoinGroupV2Response,
                LeaderAndIsrV0Response,
                LeaderAndIsrV1Response,
                LeaveGroupV0Response,
                LeaveGroupV1Response,
                ListGroupsV0Response,
                ListGroupsV1Response,
                ListOffsetV0Response,
                ListOffsetV1Response,
                ListOffsetV2Response,
                MemberAssignmentV0,
                MetadataV0Response,
                MetadataV1Response,
                MetadataV2Response,
                MetadataV3Response,
                MetadataV4Response,
                MetadataV5Response,
                OffsetCommitV0Response,
                OffsetCommitV1Response,
                OffsetCommitV2Response,
                OffsetCommitV3Response,
                OffsetFetchV0Response,
                OffsetFetchV1Response,
                OffsetFetchV2Response,
                OffsetFetchV3Response,
                OffsetForLeaderEpochV0Response,
                ProduceV0Response,
                ProduceV1Response,
                ProduceV2Response,
                SaslAuthenticateV0Response,
                SaslHandshakeV0Response,
                SaslHandshakeV1Response,
                StopReplicaV0Response,
                SyncGroupV0Response,
                SyncGroupV1Response,
                UpdateMetadataV0Response,
                UpdateMetadataV1Response,
                UpdateMetadataV2Response,
                UpdateMetadataV3Response,
                UpdateMetadataV4Response,
        ):
            obj = klass({})
            try:
                validate_schema(obj.schema)
            except (KeyError, TypeError) as e:
                self.fail("Failed to validate schema for {0}: {1}".format(klass.__name__, e))



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.find_coordinator_v1 import FindCoordinatorV1Request


class FindCoordinatorV1RequestTests(unittest.TestCase):
    def test_process_arguments(self):
        val = FindCoordinatorV1Request.process_arguments(['group', 'groupname'])
        assert val == {'coordinator_key': 'groupname', 'coordinator_type': 0}

    def test_process_arguments_missing(self):
        self.assertRaises(ArgumentError, FindCoordinatorV1Request.process_arguments, ['groupname'])

    def test_process_arguments_badtype(self):
        self.assertRaises(ArgumentError, FindCoordinatorV1Request.process_arguments, ['topic', 'groupname'])

    def test_process_arguments_toomany(self):
        self.assertRaises(ArgumentError, FindCoordinatorV1Request.process_arguments, ['group', 'groupname', 'anothergroup'])



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.list_offset_v1 import _parse_next_topic, ListOffsetV1Request


class ListOffsetV1RequestTest(unittest.TestCase):
    def test_parse_next_topic(self):
        val, rest = _parse_next_topic(['topicname', '4,2'])
        assert val == {'topic': 'topicname', 'partitions': [{'partition': 4, 'timestamp': 2}]}
        assert rest == []

    def test_parse_next_topic_multiple(self):
        val, rest = _parse_next_topic(['topicname', '4,2', '9,3'])
        assert val == {'topic': 'topicname', 'partitions': [{'partition': 4, 'timestamp': 2},
                                                            {'partition': 9, 'timestamp': 3}]}
        assert rest == []

    def test_parse_next_topic_remainder(self):
        val, rest = _parse_next_topic(['topicname', '4,2', 'nexttopic', '9,3'])
        assert val == {'topic': 'topicname', 'partitions': [{'partition': 4, 'timestamp': 2}]}
        assert rest == ['nexttopic', '9,3']

    def test_parse_next_topic_nopartitions(self):
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname'])

    def test_parse_next_topic_short_partitions(self):
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname', '4'])

    def test_parse_next_topic_extra_partitions(self):
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname', '4,2,8'])

    def test_parse_next_topic_nonnumeric(self):
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname', 'foo,2'])
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname', '4,foo'])

    def test_process_arguments(self):
        val = ListOffsetV1Request.process_arguments(['-1', 'topicname', '4,2', 'nexttopic', '9,3'])
        assert val == {'replica_id': -1,
                       'topics': [{'topic': 'topicname', 'partitions': [{'partition': 4, 'timestamp': 2}]},
                                  {'topic': 'nexttopic', 'partitions': [{'partition': 9, 'timestamp': 3}]}]}

    def test_process_arguments_nonnumeric(self):
        self.assertRaises(ArgumentError, ListOffsetV1Request.process_arguments, ['foo', 'topicname', '4,2'])

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, ListOffsetV1Request.process_arguments, ['-1', 'topicname'])



import unittest

from kafka.tools.protocol.responses import _decode_plain_type, _decode_sequence, _decode_array
from kafka.tools.protocol.types.bytebuffer import ByteBuffer


class BaseResponseTests(unittest.TestCase):
    def test_decode_plain_value(self):
        assert _decode_plain_type('int8', ByteBuffer(bytearray(b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00'))) == 1
        assert _decode_plain_type('int16', ByteBuffer(bytearray(b'\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00'))) == 1
        assert _decode_plain_type('int32', ByteBuffer(bytearray(b'\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00'))) == 1
        assert _decode_plain_type('int64', ByteBuffer(bytearray(b'\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00'))) == 1
        assert _decode_plain_type('boolean', ByteBuffer(bytearray(b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00'))) is True
        assert _decode_plain_type('string', ByteBuffer(bytearray(b'\x00\x03foo\x00\x00\x00\x00\x00'))) == 'foo'
        assert _decode_plain_type('string', ByteBuffer(bytearray(b'\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00'))) is None
        assert _decode_plain_type('bytes', ByteBuffer(bytearray(b'\x00\x00\x00\x02\x78\x23\x00\x00\x00\x00'))) == b'\x78\x23'
        assert _decode_plain_type('bytes', ByteBuffer(bytearray(b'\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00'))) is None

    def test_decode_plain_value_errors(self):
        self.assertRaises(NotImplementedError, _decode_plain_type, 'unknowntype', ByteBuffer(10))

    def test_decode_array(self):
        assert _decode_array('int8', ByteBuffer(bytearray(b'\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00'))) == [1]
        assert _decode_array('int8', ByteBuffer(bytearray(b'\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00'))) is None
        assert _decode_array([{'name': 'bar', 'type': 'int8'}], ByteBuffer(bytearray(b'\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00'))) == [{'bar': 1}]

    def test_decode_sequence(self):
        schema = [{'name': 'bar', 'type': 'int8'}]
        assert _decode_sequence(schema, ByteBuffer(bytearray(b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00'))) == {'bar': 1}

        schema = [{'name': 'bar', 'type': 'array', 'item_type': 'int8'}]
        assert _decode_sequence(schema, ByteBuffer(bytearray(b'\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00'))) == {'bar': [1]}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import six
import struct


class ByteBuffer(object):
    struct = {'b': {'pack': struct.Struct('>b').pack_into, 'unpack': struct.Struct('>b').unpack_from, 'num_bytes': 1},
              'h': {'pack': struct.Struct('>h').pack_into, 'unpack': struct.Struct('>h').unpack_from, 'num_bytes': 2},
              'i': {'pack': struct.Struct('>i').pack_into, 'unpack': struct.Struct('>i').unpack_from, 'num_bytes': 4},
              'q': {'pack': struct.Struct('>q').pack_into, 'unpack': struct.Struct('>q').unpack_from, 'num_bytes': 8}}

    def __init__(self, value):
        if isinstance(value, bytearray):
            self._buffer = value
        elif isinstance(value, six.binary_type) or isinstance(value, six.integer_types):
            self._buffer = bytearray(value)
        else:
            raise TypeError("initialization of ByteBuffer must be a bytearray, binary string, or integer")

        self._position = 0
        self._first = 0
        self._last = len(self._buffer) - 1

    @property
    def position(self):
        """The current position within the buffer"""
        return self._position

    @position.setter
    def position(self, value):
        if not isinstance(value, six.integer_types):
            raise TypeError("position must be an integer")
        self._check_position_value(value)
        self._position = value

    @property
    def limit(self):
        """The last position in the buffer"""
        return self._last

    @limit.setter
    def limit(self, value):
        if not isinstance(value, six.integer_types):
            raise TypeError("limit must be an integer")
        self._check_position_value(value, pos_type='limit')
        self._last = value

    @property
    def capacity(self):
        return self._last - self._first + 1

    @property
    def remaining(self):
        return self._last - self._position + 1

    def rewind(self):
        self._position = self._first

    def duplicate(self):
        new_bb = ByteBuffer(self._buffer)
        new_bb._first = self._first
        new_bb._position = self._position
        new_bb._last = self._last
        return new_bb

    def slice(self):
        new_bb = ByteBuffer(self._buffer)
        new_bb._first = self._position
        new_bb._position = self._position
        new_bb._last = self._last
        return new_bb

    def getInt8(self, position=None):
        return self._read_integer(position, 'b')

    def putInt8(self, value, position=None):
        return self._write_integer(value, position, 'b')

    def getInt16(self, position=None):
        return self._read_integer(position, 'h')

    def putInt16(self, value, position=None):
        return self._write_integer(value, position, 'h')

    def getInt32(self, position=None):
        return self._read_integer(position, 'i')

    def putInt32(self, value, position=None):
        return self._write_integer(value, position, 'i')

    def getInt64(self, position=None):
        return self._read_integer(position, 'q')

    def putInt64(self, value, position=None):
        return self._write_integer(value, position, 'q')

    def get(self, num_bytes, position=None):
        pos = self._get_and_check_position(position, num_bytes)
        return self._buffer[pos:pos+num_bytes]

    def put(self, byte_str, position=None):
        if not isinstance(byte_str, six.binary_type):
            raise TypeError("argument must be a binary string")

        num_bytes = len(byte_str)
        pos = self._get_and_check_position(position, num_bytes)
        self._buffer[pos:pos+num_bytes] = byte_str

    def _read_integer(self, position, struct_char):
        pos = self._get_and_check_position(position, self.struct[struct_char]['num_bytes'])
        return self.struct[struct_char]['unpack'](self._buffer, pos)[0]

    def _write_integer(self, value, position, struct_char):
        pos = self._get_and_check_position(position, self.struct[struct_char]['num_bytes'])
        # self.put(self.packers[struct_char]['func'](value), position=pos)
        self.struct[struct_char]['pack'](self._buffer, pos, value)

    def _get_and_check_position(self, position, num_bytes):
        if position is None:
            pos = self._position
            self._position = min(self._position + num_bytes, self._last + 1)
        else:
            self._check_position_value(position)
            pos = position

        if (self._last - pos + 1) < num_bytes:
            raise EOFError("not enough data left in buffer")
        return pos

    def _check_position_value(self, position, pos_type='position'):
        if (position < self._first) or (position > self._last):
            raise IndexError("{0} is out of range".format(pos_type))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import abc
import collections
import six
import struct


class ArgumentError(Exception):
    pass


def _evaluate_plain_value(value, value_type):
    if not isinstance(value_type, six.string_types):
        _evaluate_sequence(value, value_type)
    elif value_type in ('int8', 'int16', 'int32', 'int64'):
        if not isinstance(value, six.integer_types):
            raise TypeError("Expected an integer, got {0} instead".format(type(value)))
    elif value_type == 'string':
        if (value is not None) and (not isinstance(value, six.string_types)):
            raise TypeError("Expected a string, got {0} instead".format(type(value)))
    elif value_type == 'bytes':
        if (value is not None) and (not isinstance(value, six.binary_type)):
            raise TypeError("Expected a binary string, got {0} instead".format(type(value)))
    elif value_type == 'boolean':
        # Everything evaluates as boolean
        pass
    else:
        raise NotImplementedError("Reference to non-implemented type in schema: {0}".format(value_type))


def _evaluate_sequence(value, schema):
    for entry in schema:
        if entry['name'] not in value:
            raise KeyError('Value is missing an entry with key "{0}"'.format(entry['name']))
        entry_schema = entry['type']
        if isinstance(entry_schema, six.string_types) and entry_schema.lower() == 'array':
            if isinstance(value[entry['name']], collections.Sequence):
                for item in value[entry['name']]:
                    _evaluate_plain_value(item, entry['item_type'])
            elif value[entry['name']] is None:
                pass
            else:
                raise TypeError("entry '{0}' is not a list".format(entry['name']))
        else:
            _evaluate_plain_value(value[entry['name']], entry_schema)


def _encode_plain_value(value, value_type, buf):
    if value_type == 'int8':
        buf.putInt8(value)
    elif value_type == 'int16':
        buf.putInt16(value)
    elif value_type == 'int32':
        buf.putInt32(value)
    elif value_type == 'int64':
        buf.putInt64(value)
    elif value_type == 'string':
        if value is None:
            buf.putInt16(-1)
        else:
            buf.putInt16(len(value))
            buf.put(struct.pack('{0}s'.format(len(value)), value.encode("utf-8")))
    elif value_type == 'bytes':
        if value is None:
            buf.putInt32(-1)
        else:
            buf.putInt32(len(value))
            buf.put(value)
    elif value_type == 'boolean':
        buf.putInt8(1 if value else 0)
    elif isinstance(value_type, collections.Sequence):
        _encode_sequence(value, value_type, buf)
    else:
        raise NotImplementedError("Reference to non-implemented type in schema: {0}".format(value_type))


def _encode_sequence(value, schema, buf):
    for entry in schema:
        entry_schema = entry['type']
        if isinstance(entry_schema, six.string_types) and entry_schema.lower() == 'array':
            if value[entry['name']] is None:
                buf.putInt32(-1)
            else:
                buf.putInt32(len(value[entry['name']]))
                for item in value[entry['name']]:
                    _encode_plain_value(item, entry['item_type'], buf)
        else:
            _encode_plain_value(value[entry['name']], entry_schema, buf)


@six.add_metaclass(abc.ABCMeta)
class BaseRequest():  # pragma: no cover
    @abc.abstractproperty
    def schema(self):
        raise NotImplementedError

    @abc.abstractproperty
    def api_key(self):
        raise NotImplementedError

    @abc.abstractproperty
    def api_version(self):
        pass

    @abc.abstractproperty
    def cmd(self):
        raise NotImplementedError

    @abc.abstractproperty
    def response(self):
        raise NotImplementedError

    @abc.abstractproperty
    def help_string(self):
        raise NotImplementedError

    # By default, requests are not used in the CLI
    supports_cli = False

    def __init__(self, value):
        _evaluate_sequence(value, self.schema)
        self._request = value

    def encode(self, buf):
        _encode_sequence(self._request, self.schema, buf)

    # This is not an abstract method because requests that are not supported in the CLI will not override it
    @classmethod
    def process_arguments(cls, cmd_args):
        """IMPORTANT: this is class method, override it with @classmethod!"""
        raise NotImplementedError

    def __hash__(self):
        return id(self)

    def __len__(self):
        return len(self._request)

    def __contains__(self, k):
        return k in self._request

    def __getitem__(self, k):
        return self._request[k]

    def __setitem__(self, k):
        raise NotImplementedError

    def __delitem__(self, k):
        raise NotImplementedError



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.offset_fetch_v0 import OffsetFetchV0Request, _parse_topic_set
from kafka.tools.protocol.responses.offset_fetch_v2 import OffsetFetchV2Response


class OffsetFetchV2Request(OffsetFetchV0Request):
    api_version = 2
    response = OffsetFetchV2Response
    cmd = "OffsetFetch"

    supports_cli = True
    help_string = ("Request:     {0}V{1}\n".format(cmd, api_version) +
                   "Format:      {0}V{1} group_id [topic(,partition ...) ...]\n".format(cmd, api_version) +
                   "Description: Fetch offsets for the specified topic partitions for the consumer group. If no\n" +
                   "             topic is specified, fetch for all topic partitions.\n" +
                   "Example:     {0}V{1} ExampleGroup ExampleTopic,0,1,2 ExampleTopic2,0\n".format(cmd, api_version) +
                   "             {0}V{1} ExampleGroup".format(cmd, api_version))

    @classmethod
    def process_arguments(cls, cmd_args):
        if len(cmd_args) == 1:
            return {'group_id': cmd_args[0], 'topics': None}
        elif len(cmd_args) == 0:
            raise ArgumentError("OffsetFetchV{0} requires at least 1 argument".format(cls.api_version))
        else:
            values = {'group_id': cmd_args[0], 'topics': []}

            for topic_set in cmd_args[1:]:
                values['topics'].append(_parse_topic_set(topic_set))

            return values



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.protocol.requests.topic_metadata_v3 import TopicMetadataV3Request
from kafka.tools.protocol.responses.metadata_v4 import MetadataV4Response


class TopicMetadataV4Request(TopicMetadataV3Request):
    api_version = 4
    response = MetadataV4Response
    cmd = "TopicMetadata"

    help_string = ("Request:     {0}V{1}\n".format(cmd, api_version) +
                   "Format:      {0}V{1} [topic_name ...]\n".format(cmd, api_version) +
                   "Description: Fetch metadata for the specified topics. If no topics are specified, all topics\n" +
                   "             are requested. Note that auto-creation is not supported.\n")

    schema = [
        {'name': 'topics', 'type': 'array', 'item_type': 'string'},
        {'name': 'allow_auto_topic_creation', 'type': 'boolean'}
    ]

    @classmethod
    def process_arguments(cls, cmd_args):
        # This looks weird, but it's correct. The list is the first item
        if len(cmd_args) == 0:
            return {'allow_auto_topic_creation': False, 'topics': None}
        else:
            return {'allow_auto_topic_creation': False, 'topics': cmd_args}



import unittest
from mock import patch


from kafka.tools.protocol.requests import _evaluate_plain_value, _evaluate_sequence, _encode_plain_value, _encode_sequence
from kafka.tools.protocol.types.bytebuffer import ByteBuffer


class BaseRequestTests(unittest.TestCase):
    def test_evaluate_plain_value(self):
        _evaluate_plain_value(1, 'int8')
        _evaluate_plain_value(1, 'int16')
        _evaluate_plain_value(1, 'int32')
        _evaluate_plain_value(1, 'int64')
        _evaluate_plain_value('foo', 'string')
        _evaluate_plain_value(b'foo', 'bytes')
        _evaluate_plain_value(True, 'boolean')
        assert True

    @patch('kafka.tools.protocol.requests._evaluate_sequence')
    def test_evaluate_plain_value_sequence(self, mock_eval_seq):
        _evaluate_plain_value({'foo': 'bar'}, [{'name': 'foo', 'type': 'string'}])
        mock_eval_seq.assert_called_once_with({'foo': 'bar'}, [{'name': 'foo', 'type': 'string'}])

    def test_evaluate_plain_value_errors(self):
        self.assertRaises(TypeError, _evaluate_plain_value, 'foo', 'int8')
        self.assertRaises(TypeError, _evaluate_plain_value, 'foo', 'int16')
        self.assertRaises(TypeError, _evaluate_plain_value, 'foo', 'int32')
        self.assertRaises(TypeError, _evaluate_plain_value, 'foo', 'int64')
        self.assertRaises(TypeError, _evaluate_plain_value, 1, 'string')
        self.assertRaises(TypeError, _evaluate_plain_value, 1, 'bytes')
        self.assertRaises(NotImplementedError, _evaluate_plain_value, 1, 'unknowntype')

    def test_evaluate_sequence(self):
        _evaluate_sequence({'foo': 1}, [{'name': 'foo', 'type': 'int8'}])
        _evaluate_sequence({'foo': None}, [{'name': 'foo', 'type': 'array', 'item_type': 'int8'}])
        _evaluate_sequence({'foo': [1]}, [{'name': 'foo', 'type': 'array', 'item_type': 'int8'}])
        _evaluate_sequence({'foo': [{'bar': 1}]}, [{'name': 'foo', 'type': 'array', 'item_type': [{'name': 'bar', 'type': 'int8'}]}])

    def test_evaluate_sequence_errors(self):
        self.assertRaises(KeyError, _evaluate_sequence, {}, [{'name': 'foo', 'type': 'int8'}])
        self.assertRaises(TypeError, _evaluate_sequence, {'foo': 1}, [{'name': 'foo', 'type': 'array', 'item_type': 'int8'}])

    def test_encode_plain_value(self):
        buf = ByteBuffer(10)
        _encode_plain_value(1, 'int8', buf)
        assert buf.get(10, 0) == bytearray(b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00')

        buf = ByteBuffer(10)
        _encode_plain_value(1, 'int16', buf)
        assert buf.get(10, 0) == bytearray(b'\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00')

        buf = ByteBuffer(10)
        _encode_plain_value(1, 'int32', buf)
        assert buf.get(10, 0) == bytearray(b'\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00')

        buf = ByteBuffer(10)
        _encode_plain_value(1, 'int64', buf)
        assert buf.get(10, 0) == bytearray(b'\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00')

        buf = ByteBuffer(10)
        _encode_plain_value(True, 'boolean', buf)
        assert buf.get(10, 0) == bytearray(b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00')

        buf = ByteBuffer(10)
        _encode_plain_value('foo', 'string', buf)
        assert buf.get(10, 0) == bytearray(b'\x00\x03foo\x00\x00\x00\x00\x00')

        buf = ByteBuffer(10)
        _encode_plain_value(None, 'string', buf)
        assert buf.get(10, 0) == bytearray(b'\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00')

        buf = ByteBuffer(10)
        _encode_plain_value(b'\x78\x23', 'bytes', buf)
        assert buf.get(10, 0) == bytearray(b'\x00\x00\x00\x02\x78\x23\x00\x00\x00\x00')

        buf = ByteBuffer(10)
        _encode_plain_value(None, 'bytes', buf)
        assert buf.get(10, 0) == bytearray(b'\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00')

    def test_encode_plain_value_errors(self):
        buf = ByteBuffer(10)
        self.assertRaises(NotImplementedError, _encode_plain_value, {'foo': 1}, {'name': 'foo', 'type': 'unknowntype'}, buf)

    def test_encode_sequence(self):
        buf = ByteBuffer(10)
        _encode_sequence({'foo': 1}, [{'name': 'foo', 'type': 'int8'}], buf)
        assert buf.get(10, 0) == bytearray(b'\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00')

        buf = ByteBuffer(10)
        _encode_sequence({'foo': None}, [{'name': 'foo', 'type': 'array', 'item_type': 'int8'}], buf)
        assert buf.get(10, 0) == bytearray(b'\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00')

        buf = ByteBuffer(10)
        _encode_sequence({'foo': [1]}, [{'name': 'foo', 'type': 'array', 'item_type': 'int8'}], buf)
        assert buf.get(10, 0) == bytearray(b'\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00')

        buf = ByteBuffer(10)
        _encode_sequence({'foo': [{'bar': 1}]}, [{'name': 'foo', 'type': 'array', 'item_type': [{'name': 'bar', 'type': 'int8'}]}], buf)
        assert buf.get(10, 0) == bytearray(b'\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00')



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.offset_fetch_v2 import OffsetFetchV2Request


class OffsetFetchV2RequestTests(unittest.TestCase):
    def test_process_arguments(self):
        val = OffsetFetchV2Request.process_arguments(['groupname', 'topicname,4', 'nexttopic,9'])
        assert val == {'group_id': 'groupname',
                       'topics': [{'topic': 'topicname', 'partitions': [4]},
                                  {'topic': 'nexttopic', 'partitions': [9]}]}

    def test_process_arguments_alltopics(self):
        val = OffsetFetchV2Request.process_arguments(['groupname'])
        assert val == {'group_id': 'groupname',
                       'topics': None}

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, OffsetFetchV2Request.process_arguments, [])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.create_topics_v0 import CreateTopicsV0Request, _parse_remaining_args
from kafka.tools.protocol.responses.create_topics_v1 import CreateTopicsV1Response


class CreateTopicsV1Request(CreateTopicsV0Request):
    api_version = 1
    cmd = "CreateTopics"
    response = CreateTopicsV1Response

    supports_cli = True
    help_string = ("Request:     {0}V{1}\n".format(cmd, api_version) +
                   "Format:      {0}V{1} dry_run timeout (topic_name,num_partitions,".format(cmd, api_version) +
                   "replication_factor[,config=value]... )...\n" +
                   "             {0}V{1} dry_run timeout (topic_name,(partition_id=".format(cmd, api_version) +
                   "replica_id[|replica_id]...)...[,config=value]... )...\n"
                   "Description: Create the topics with the specified values. Note that EITHER num_partitions\n" +
                   "             and replication_factor, OR replica_assignment, MUST be provided for each topic\n" +
                   "             but not both. dry_run can only be 'true' or 'false'.\n" +
                   "Examples:    {0}V{1} true 30 ExampleTopicName,4,2 OtherExampleTopic,4,2\n".format(cmd, api_version) +
                   "             {0}V{1} false 30 ExampleTopicName,4,2,retention.ms=3600000\n".format(cmd, api_version) +
                   "             {0}V{1} false 30 ExampleTopicName,0=1|2,1=2|3,2=3|1,3=1|3\n".format(cmd, api_version))

    schema = [
        {'name': 'topics',
         'type': 'array',
         'item_type': [
             {'name': 'topic', 'type': 'string'},
             {'name': 'num_partitions', 'type': 'int32'},
             {'name': 'replication_factor', 'type': 'int16'},
             {'name': 'replica_assignment',
              'type': 'array',
              'item_type': [
                  {'name': 'partition_id', 'type': 'int32'},
                  {'name': 'replicas', 'type': 'array', 'item_type': 'int32'},
              ]},
             {'name': 'configs',
              'type': 'array',
              'item_type': [
                  {'name': 'config_key', 'type': 'string'},
                  {'name': 'config_value', 'type': 'string'},
              ]},
         ]},
        {'name': 'timeout', 'type': 'int32'},
        {'name': 'validate_only', 'type': 'boolean'},
    ]

    @classmethod
    def process_arguments(cls, cmd_args):
        if len(cmd_args) < 3:
            raise ArgumentError("At least one topic definition must be provided")
        if not cmd_args[0] in ('true', 'false'):
            raise ArgumentError("The first argument may only be true or false")
        if not cmd_args[1].isdigit():
            raise ArgumentError("The second argument must be an integer")

        values = {'validate_only': cmd_args[0] == 'true', 'topics': [], 'timeout': int(cmd_args[1])}
        for topic_def in cmd_args[2:]:
            tparts = topic_def.split(",")
            topic = {'topic': tparts[0],
                     'num_partitions': -1,
                     'replication_factor': -1,
                     'replica_assignment': [],
                     'configs': []}

            _parse_remaining_args(topic, tparts[1:])
            values['topics'].append(topic)

        return values



import unittest
from mock import patch

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.create_topics_v1 import CreateTopicsV1Request


class CreateTopicsV1RequestTests(unittest.TestCase):
    def setUp(self):
        self.topic = {'topic': 'topicname',
                      'num_partitions': -1,
                      'replication_factor': -1,
                      'replica_assignment': [],
                      'configs': []}

    @patch('kafka.tools.protocol.requests.create_topics_v1._parse_remaining_args')
    def test_process_arguments(self, mock_parse):
        val = CreateTopicsV1Request.process_arguments(['true', '3296', 'topicname,someargs'])
        mock_parse.assert_called_once_with(self.topic, ['someargs'])
        assert val == {'timeout': 3296, 'topics': [self.topic], 'validate_only': True}

    def test_process_arguments_nodryrun(self):
        self.assertRaises(ArgumentError, CreateTopicsV1Request.process_arguments, ['3296', 'topicname,someargs', 'anothertopic,moreargs'])

    def test_process_arguments_notimeout(self):
        self.assertRaises(ArgumentError, CreateTopicsV1Request.process_arguments, ['true', 'topicname,someargs', 'anothertopic,moreargs'])

    def test_process_arguments_notopic(self):
        self.assertRaises(ArgumentError, CreateTopicsV1Request.process_arguments, ['true', '3296'])



import collections
import six


def _validate_schema_entry_type(entry_type):
    if isinstance(entry_type, six.string_types):
        if entry_type.lower() not in ('int8', 'int16', 'int32', 'int64', 'string', 'bytes', 'boolean', 'array'):
            raise TypeError('Unknown schema data type: {0}'.format(entry_type))
    elif isinstance(entry_type, collections.Sequence):
        validate_schema(entry_type)
    else:
        raise TypeError('Schema type must be a string type or a list, not {0}'.format(type(entry_type)))


def _validate_schema_entry(entry):
    if not isinstance(entry, dict):
        raise TypeError('Expected schema entries to be dicts, not {0}'.format(type(entry)))
    if not all(k in entry for k in ('name', 'type')):
        raise KeyError('Schema entries must have both name and type keys')
    if not isinstance(entry['name'], six.string_types):
        raise TypeError('Schema entry name must be a string, not {0}'.format(type(entry['name'])))


# A method for use in tests to validate that a schema (for requests or responses) is properly formed
def validate_schema(schema):
    for entry in schema:
        _validate_schema_entry(entry)
        if isinstance(entry['type'], six.string_types) and (entry['type'].lower() == 'array'):
            if 'item_type' not in entry:
                raise KeyError('Array schema entries must have an item_type key')
            _validate_schema_entry_type(entry['item_type'])
        else:
            _validate_schema_entry_type(entry['type'])



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.offset_commit_v2 import OffsetCommitV2Request


class OffsetCommitV2RequestTests(unittest.TestCase):
    def test_process_arguments(self):
        val = OffsetCommitV2Request.process_arguments(['groupname', '16', 'membername', '76', 'topicname', '4,2', 'nexttopic', '9,3'])
        assert val == {'group_id': 'groupname',
                       'group_generation_id': 16,
                       'member_id': 'membername',
                       'retention_time': 76,
                       'topics': [{'topic': 'topicname', 'partitions': [{'partition': 4, 'offset': 2, 'metadata': None}]},
                                  {'topic': 'nexttopic', 'partitions': [{'partition': 9, 'offset': 3, 'metadata': None}]}]}

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, OffsetCommitV2Request.process_arguments, ['groupname', '16', 'membername', '76', 'topicname'])

    def test_process_arguments_nonnumeric(self):
        self.assertRaises(ArgumentError, OffsetCommitV2Request.process_arguments, ['groupname', 'foo', 'membername', '76', 'topicname', '4,2'])
        self.assertRaises(ArgumentError, OffsetCommitV2Request.process_arguments, ['groupname', '16', 'membername', 'foo', 'topicname', '4,2'])



import unittest
from mock import patch

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.join_group_v1 import JoinGroupV1Request


class JoinGroupV0RequestTests(unittest.TestCase):
    @patch('kafka.tools.protocol.requests.join_group_v1._parse_group_protocol')
    def test_process_arguments(self, mock_parse):
        mock_parse.return_value = {'protocol_name': 'protocoltype', 'protocol_metadata': b'\x89\x12\x34\xab\xdb'}
        val = JoinGroupV1Request.process_arguments(['groupname', '3296', '89243', 'membername',
                                                    'protocolname', 'somegroupprotocols'])

        mock_parse.assert_called_once_with('somegroupprotocols')
        assert val == {'group_id': 'groupname',
                       'session_timeout': 3296,
                       'rebalance_timeout': 89243,
                       'member_id': 'membername',
                       'protocol_type': 'protocolname',
                       'group_protocols': [{'protocol_name': 'protocoltype', 'protocol_metadata': b'\x89\x12\x34\xab\xdb'}]}

    def test_process_arguments_missing(self):
        self.assertRaises(ArgumentError, JoinGroupV1Request.process_arguments, [])

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, JoinGroupV1Request.process_arguments, ['foo', '23412', '89243', 'bar'])

    def test_process_arguments_nonnumeric(self):
        self.assertRaises(ArgumentError, JoinGroupV1Request.process_arguments, ['groupname', 'foo', '89243',
                                                                                'membername', 'protocolname',
                                                                                'somegroupprotocols'])



import collections
import unittest
import socket
from mock import MagicMock, call, patch

from kafka.tools.configuration import ClientConfiguration
from kafka.tools.exceptions import ConfigurationException, ConnectionError
from kafka.tools.models.broker import Broker
from kafka.tools.models.topic import Topic
from kafka.tools.protocol.requests.api_versions_v0 import ApiVersionsV0Request
from kafka.tools.protocol.responses.api_versions_v0 import ApiVersionsV0Response
from kafka.tools.protocol.types.bytebuffer import ByteBuffer


class BrokerTests(unittest.TestCase):
    def setUp(self):
        self.configuration = ClientConfiguration(num_retries=3, retry_backoff=0.5)
        self.mock_sock = MagicMock()
        self.broker = Broker('brokerhost1.example.com', id=1, sock=self.mock_sock, configuration=self.configuration)

    def add_partitions(self, pos, num):
        topic = Topic('testTopic', num)
        self.broker.partitions[pos] = []
        for i in range(num):
            self.broker.partitions[pos].append(topic.partitions[i])

    def test_broker_create(self):
        assert self.broker.id == 1
        assert self.broker.hostname == 'brokerhost1.example.com'
        assert self.broker.partitions == {}
        assert isinstance(self.broker._sock, MagicMock)

    def test_broker_create_from_json_basic(self):
        jsonstr = '{"jmx_port":-1,"timestamp":"1466985807242","endpoints":["PLAINTEXT://10.0.0.10:9092"],"host":"10.0.0.10","version":3,"port":9092}'
        broker2 = Broker.create_from_json(1, jsonstr)
        assert broker2.id == 1
        assert broker2.hostname == '10.0.0.10'
        assert broker2.partitions == {}

    def test_broker_get_endpoint(self):
        jsonstr = ('{"jmx_port":-1,"timestamp":"1466985807242","endpoints":["PLAINTEXT://10.0.0.10:9092", "SSL://10.0.0.10:2834"],'
                   '"host":"10.0.0.10","version":3,"port":9092}')
        broker = Broker.create_from_json(1, jsonstr)
        endpoint = broker.get_endpoint("SSL")
        assert endpoint.protocol == "SSL"
        assert endpoint.hostname == "10.0.0.10"
        assert endpoint.port == 2834

    def test_broker_create_from_json_extended(self):
        jsonstr = '{"jmx_port":-1,"timestamp":"1466985807242","endpoints":["PLAINTEXT://10.0.0.10:9092"],"host":"10.0.0.10","version":3,"port":9092}'
        broker2 = Broker.create_from_json(1, jsonstr)
        assert broker2.jmx_port == -1
        assert broker2.timestamp == "1466985807242"
        assert len(broker2.endpoints) == 1
        assert broker2.endpoints[0].protocol == 'PLAINTEXT'
        assert broker2.endpoints[0].hostname == '10.0.0.10'
        assert broker2.endpoints[0].port == 9092
        assert broker2.version == 3
        assert broker2.port == 9092

    def test_broker_create_from_json_bad_host(self):
        jsonstr = '{"jmx_port":-1,"timestamp":"1466985807242","endpoints":["PLAINTEXT://10.0.0.10:9092"],"hostname":"10.0.0.10","version":3,"port":9092}'
        self.assertRaises(ConfigurationException, Broker.create_from_json, 1, jsonstr)

    def test_broker_create_from_json_bad_jmx_port(self):
        jsonstr = '{"timestamp":"1466985807242","endpoints":["PLAINTEXT://10.0.0.10:9092"],"host":"10.0.0.10","version":3,"port":9092}'
        broker2 = Broker.create_from_json(1, jsonstr)
        assert broker2.hostname == '10.0.0.10'

    def test_broker_copy_without_partitions(self):
        broker2 = self.broker.copy()
        assert broker2.id == 1
        assert broker2.hostname == 'brokerhost1.example.com'
        assert self.broker is not broker2

    def test_broker_copy_with_partitions(self):
        self.add_partitions(0, 1)
        broker2 = self.broker.copy()
        assert broker2.id == 1
        assert broker2.hostname == 'brokerhost1.example.com'
        assert broker2.partitions == {}
        assert self.broker is not broker2

    def test_broker_num_leaders(self):
        self.add_partitions(0, 2)
        assert self.broker.num_leaders() == 2

    def test_broker_num_partitions_at_position_zero(self):
        self.add_partitions(1, 2)
        assert self.broker.num_partitions_at_position(0) == 0

    def test_broker_num_partitions_single_position(self):
        self.add_partitions(0, 2)
        assert self.broker.num_partitions() == 2

    def test_broker_num_partitions_two_position(self):
        self.add_partitions(0, 2)
        self.add_partitions(1, 1)
        assert self.broker.num_partitions() == 3

    def test_broker_percent_leaders(self):
        self.add_partitions(0, 2)
        self.add_partitions(1, 2)
        assert self.broker.percent_leaders() == 50.0

    def test_broker_percent_leaders_zero(self):
        self.add_partitions(1, 1)
        assert self.broker.percent_leaders() == 0.0

    def test_broker_percent_leaders_no_partitions(self):
        assert self.broker.percent_leaders() == 0.0

    def test_broker_equality(self):
        broker2 = Broker('brokerhost1.example.com', id=1)
        assert self.broker == broker2

    def test_broker_inequality_hostname(self):
        broker2 = Broker('brokerhost2.example.com', id=1)
        assert self.broker != broker2

    def test_broker_inequality_id(self):
        broker2 = Broker('brokerhost1.example.com', id=2)
        assert self.broker != broker2

    def test_broker_equality_typeerror(self):
        self.assertRaises(TypeError, self.broker.__eq__, None)

    def test_broker_get_socket(self):
        sock = self.broker._get_socket(None)
        assert isinstance(sock, socket.socket)

    @patch('kafka.tools.models.broker.socket.socket')
    def test_broker_get_socket_wrapped(self, mock_socket):
        mock_context = MagicMock()
        mock_context.wrap_socket.return_value = 'fakewrappedsocket'
        mock_socket.return_value = 'fakesocket'

        sock = self.broker._get_socket(mock_context)
        mock_context.wrap_socket.assert_called_once_with('fakesocket', server_hostname='brokerhost1.example.com')
        print(sock)
        assert sock == 'fakewrappedsocket'

    def test_broker_connect(self):
        self.broker.connect()
        self.mock_sock.connect.assert_called_once_with(('brokerhost1.example.com', 9092))

    def test_broker_connect_error(self):
        self.mock_sock.connect.side_effect = socket.error
        self.assertRaises(ConnectionError, self.broker.connect)

    def test_broker_close(self):
        self.broker.close()
        self.mock_sock.shutdown.assert_called_once()
        self.mock_sock.close.assert_called_once()

    def test_broker_close_not_open(self):
        self.mock_sock.shutdown.side_effect = OSError()

        self.broker.close()
        self.mock_sock.close.assert_called_once()

    @patch.object(Broker, '_read_bytes')
    def test_broker_single_send(self, mock_read_bytes):
        # recv is called to get the response size (first 4 bytes)
        self.mock_sock.recv.return_value = b'\x00\x00\x00\x10'

        # read_bytes returns the response payload, minus the response size
        # correlation_id (4), error (2), array of (api_key (2), min_version (2), max_version (2))
        mock_read_bytes.return_value = b'\x00\x00\x00\x01\x00\x00\x00\x00\x00\x01\x00\x01\x01\x01\x02\x02'

        request = ApiVersionsV0Request({})
        (correlation_id, response) = self.broker._single_send(request)

        # Check that the request was encoded properly
        self.mock_sock.sendall.assert_called_once_with(bytearray(b'\x00\x00\x00\x15\x00\x12\x00\x00\x00\x00\x00\x01\x00\x0bkafka-tools'))
        self.mock_sock.recv.assert_called_once_with(4)
        mock_read_bytes.assert_called_once_with(16)

        # Check that the response is what was expected to be decoded
        assert isinstance(response, ApiVersionsV0Response)
        assert response.correlation_id == 1
        assert response['error'] == 0
        assert isinstance(response['api_versions'], collections.Sequence)
        assert len(response['api_versions']) == 1
        assert len(response['api_versions'][0]) == 3
        assert response['api_versions'][0]['api_key'] == 1
        assert response['api_versions'][0]['min_version'] == 257
        assert response['api_versions'][0]['max_version'] == 514

        # Correlation ID must be incremented after each request
        assert self.broker._correlation_id == 2

    @patch.object(Broker, '_read_bytes')
    def test_broker_single_send_error(self, mock_read_bytes):
        self.mock_sock.recv.side_effect = socket.error
        mock_read_bytes.return_value = b'\x00\x00\x00\x01\x00\x00\x00\x00\x00\x01\x00\x01\x01\x01\x02\x02'

        request = ApiVersionsV0Request({})
        self.assertRaises(ConnectionError, self.broker._single_send, request)

    @patch.object(ByteBuffer, 'getInt32')
    def test_broker_single_send_short(self, mock_getint):
        self.mock_sock.recv.return_value = b'\x00\x00\x00\x10'
        mock_getint.side_effect = EOFError

        request = ApiVersionsV0Request({})
        self.assertRaises(ConnectionError, self.broker._single_send, request)

    @patch.object(Broker, '_single_send')
    @patch.object(Broker, 'connect')
    @patch.object(Broker, 'close')
    def test_broker_send(self, mock_close, mock_connect, mock_send):
        mock_send.return_value = 'fakeresponse'
        self.broker.send('fakerequest')
        mock_connect.assert_not_called()
        mock_close.assert_not_called()
        mock_send.assert_called_once_with('fakerequest')

    @patch.object(Broker, '_single_send')
    @patch.object(Broker, 'connect')
    @patch.object(Broker, 'close')
    def test_broker_send_retry(self, mock_close, mock_connect, mock_send):
        def close_broker():
            self.broker._sock = None

        mock_send.side_effect = [ConnectionError, 'fakeresponse']
        mock_close.side_effect = close_broker
        self.broker.send('fakerequest')
        mock_close.assert_called_once_with()
        mock_connect.assert_called_once_with()
        mock_send.assert_has_calls([call('fakerequest'), call('fakerequest')])

    @patch.object(Broker, '_single_send')
    @patch.object(Broker, 'connect')
    @patch.object(Broker, 'close')
    def test_broker_send_exhaust_retry(self, mock_close, mock_connect, mock_send):
        def close_broker():
            self.broker._sock = None

        mock_send.side_effect = ConnectionError
        mock_close.side_effect = close_broker
        self.assertRaises(ConnectionError, self.broker.send, 'fakerequest')
        mock_close.assert_has_calls([call(), call()])
        mock_connect.assert_has_calls([call(), call()])
        mock_send.assert_has_calls([call('fakerequest'), call('fakerequest'), call('fakerequest')])

    def test_broker_read_bytes(self):
        self.mock_sock.recv.return_value = b'\x01\x02\x03\x04'
        response = self.broker._read_bytes(4)

        self.mock_sock.recv.assert_called_once_with(4)
        assert response == b'\x01\x02\x03\x04'

    def test_broker_read_bytes_concat(self):
        self.mock_sock.recv.side_effect = [b'\x01\x02', b'\x03\x04']
        response = self.broker._read_bytes(4)

        self.mock_sock.recv.assert_has_calls([call(4), call(2)])
        assert response == b'\x01\x02\x03\x04'

    def test_broker_read_bytes_nodata(self):
        self.mock_sock.recv.return_value = b''
        self.assertRaises(socket.error, self.broker._read_bytes, 4)

    def test_broker_read_bytes_error(self):
        self.mock_sock.recv.side_effect = socket.error
        self.assertRaises(socket.error, self.broker._read_bytes, 4)



import unittest
from tests.tools.protocol.utilities import validate_schema

from kafka.tools.protocol.requests.alter_configs_v0 import AlterConfigsV0Request
from kafka.tools.protocol.requests.alter_replica_log_dirs_v0 import AlterReplicaLogDirsV0Request
from kafka.tools.protocol.requests.api_versions_v0 import ApiVersionsV0Request
from kafka.tools.protocol.requests.api_versions_v1 import ApiVersionsV1Request
from kafka.tools.protocol.requests.controlled_shutdown_v0 import ControlledShutdownV0Request
from kafka.tools.protocol.requests.controlled_shutdown_v1 import ControlledShutdownV1Request
from kafka.tools.protocol.requests.create_acls_v0 import CreateAclsV0Request
from kafka.tools.protocol.requests.create_partitions_v0 import CreatePartitionsV0Request
from kafka.tools.protocol.requests.create_topics_v0 import CreateTopicsV0Request
from kafka.tools.protocol.requests.create_topics_v1 import CreateTopicsV1Request
from kafka.tools.protocol.requests.create_topics_v2 import CreateTopicsV2Request
from kafka.tools.protocol.requests.delete_acls_v0 import DeleteAclsV0Request
from kafka.tools.protocol.requests.delete_records_v0 import DeleteRecordsV0Request
from kafka.tools.protocol.requests.delete_topics_v0 import DeleteTopicsV0Request
from kafka.tools.protocol.requests.delete_topics_v1 import DeleteTopicsV1Request
from kafka.tools.protocol.requests.describe_acls_v0 import DescribeAclsV0Request
from kafka.tools.protocol.requests.describe_configs_v0 import DescribeConfigsV0Request
from kafka.tools.protocol.requests.describe_groups_v0 import DescribeGroupsV0Request
from kafka.tools.protocol.requests.describe_groups_v1 import DescribeGroupsV1Request
from kafka.tools.protocol.requests.describe_log_dirs_v0 import DescribeLogDirsV0Request
from kafka.tools.protocol.requests.find_coordinator_v1 import FindCoordinatorV1Request
from kafka.tools.protocol.requests.group_coordinator_v0 import GroupCoordinatorV0Request
from kafka.tools.protocol.requests.heartbeat_v0 import HeartbeatV0Request
from kafka.tools.protocol.requests.join_group_v0 import JoinGroupV0Request
from kafka.tools.protocol.requests.join_group_v1 import JoinGroupV1Request
from kafka.tools.protocol.requests.join_group_v2 import JoinGroupV2Request
from kafka.tools.protocol.requests.leader_and_isr_v0 import LeaderAndIsrV0Request
from kafka.tools.protocol.requests.leader_and_isr_v1 import LeaderAndIsrV1Request
from kafka.tools.protocol.requests.leave_group_v0 import LeaveGroupV0Request
from kafka.tools.protocol.requests.leave_group_v1 import LeaveGroupV1Request
from kafka.tools.protocol.requests.list_groups_v0 import ListGroupsV0Request
from kafka.tools.protocol.requests.list_groups_v1 import ListGroupsV1Request
from kafka.tools.protocol.requests.list_offset_v0 import ListOffsetV0Request
from kafka.tools.protocol.requests.list_offset_v1 import ListOffsetV1Request
from kafka.tools.protocol.requests.list_offset_v2 import ListOffsetV2Request
from kafka.tools.protocol.requests.offset_commit_v0 import OffsetCommitV0Request
from kafka.tools.protocol.requests.offset_commit_v1 import OffsetCommitV1Request
from kafka.tools.protocol.requests.offset_commit_v2 import OffsetCommitV2Request
from kafka.tools.protocol.requests.offset_commit_v3 import OffsetCommitV3Request
from kafka.tools.protocol.requests.offset_fetch_v0 import OffsetFetchV0Request
from kafka.tools.protocol.requests.offset_fetch_v1 import OffsetFetchV1Request
from kafka.tools.protocol.requests.offset_fetch_v2 import OffsetFetchV2Request
from kafka.tools.protocol.requests.offset_fetch_v3 import OffsetFetchV3Request
from kafka.tools.protocol.requests.offsets_for_leader_epoch_v0 import OffsetForLeaderEpochV0Request
from kafka.tools.protocol.requests.sasl_handshake_v0 import SaslHandshakeV0Request
from kafka.tools.protocol.requests.sasl_handshake_v1 import SaslHandshakeV1Request
from kafka.tools.protocol.requests.sasl_authenticate_v0 import SaslAuthenticateV0Request
from kafka.tools.protocol.requests.stop_replica_v0 import StopReplicaV0Request
from kafka.tools.protocol.requests.sync_group_v0 import SyncGroupV0Request
from kafka.tools.protocol.requests.sync_group_v1 import SyncGroupV1Request
from kafka.tools.protocol.requests.topic_metadata_v0 import TopicMetadataV0Request
from kafka.tools.protocol.requests.topic_metadata_v1 import TopicMetadataV1Request
from kafka.tools.protocol.requests.topic_metadata_v2 import TopicMetadataV2Request
from kafka.tools.protocol.requests.topic_metadata_v3 import TopicMetadataV3Request
from kafka.tools.protocol.requests.topic_metadata_v4 import TopicMetadataV4Request
from kafka.tools.protocol.requests.topic_metadata_v5 import TopicMetadataV5Request
from kafka.tools.protocol.requests.update_metadata_v0 import UpdateMetadataV0Request
from kafka.tools.protocol.requests.update_metadata_v1 import UpdateMetadataV1Request
from kafka.tools.protocol.requests.update_metadata_v2 import UpdateMetadataV2Request
from kafka.tools.protocol.requests.update_metadata_v3 import UpdateMetadataV3Request
from kafka.tools.protocol.requests.update_metadata_v4 import UpdateMetadataV4Request


class ValidateResponseSchemaTests(unittest.TestCase):
    def test_schema_validation(self):
        for klass in (
            AlterConfigsV0Request,
            AlterReplicaLogDirsV0Request,
            ApiVersionsV0Request,
            ApiVersionsV1Request,
            ControlledShutdownV0Request,
            ControlledShutdownV1Request,
            CreateAclsV0Request,
            CreatePartitionsV0Request,
            CreateTopicsV0Request,
            CreateTopicsV1Request,
            CreateTopicsV2Request,
            DeleteAclsV0Request,
            DeleteRecordsV0Request,
            DeleteTopicsV0Request,
            DeleteTopicsV1Request,
            DescribeAclsV0Request,
            DescribeConfigsV0Request,
            DescribeGroupsV0Request,
            DescribeGroupsV1Request,
            DescribeLogDirsV0Request,
            FindCoordinatorV1Request,
            GroupCoordinatorV0Request,
            HeartbeatV0Request,
            JoinGroupV0Request,
            JoinGroupV1Request,
            JoinGroupV2Request,
            LeaderAndIsrV0Request,
            LeaderAndIsrV1Request,
            LeaveGroupV0Request,
            LeaveGroupV1Request,
            ListGroupsV0Request,
            ListGroupsV1Request,
            ListOffsetV0Request,
            ListOffsetV1Request,
            ListOffsetV2Request,
            OffsetCommitV0Request,
            OffsetCommitV1Request,
            OffsetCommitV2Request,
            OffsetCommitV3Request,
            OffsetFetchV0Request,
            OffsetFetchV1Request,
            OffsetFetchV2Request,
            OffsetFetchV3Request,
            OffsetForLeaderEpochV0Request,
            SaslHandshakeV0Request,
            SaslHandshakeV1Request,
            SaslAuthenticateV0Request,
            StopReplicaV0Request,
            SyncGroupV0Request,
            SyncGroupV1Request,
            TopicMetadataV0Request,
            TopicMetadataV1Request,
            TopicMetadataV2Request,
            TopicMetadataV3Request,
            TopicMetadataV4Request,
            TopicMetadataV5Request,
            UpdateMetadataV0Request,
            UpdateMetadataV1Request,
            UpdateMetadataV2Request,
            UpdateMetadataV3Request,
            UpdateMetadataV4Request,
        ):
            try:
                validate_schema(klass.schema)
            except (KeyError, TypeError) as e:
                self.fail("Failed to validate schema for {0}: {1}".format(klass.__name__, e))



import unittest

from kafka.tools.protocol.types.bytebuffer import ByteBuffer


class ByteBufferTests(unittest.TestCase):
    def setUp(self):
        self.test_bb = ByteBuffer(b'\x03\x05\x07\x09\x11\x13\x15\x17\x19\x21')

    def test_create_allocate(self):
        val = ByteBuffer(4)
        assert val.position == 0
        assert val.capacity == 4
        assert val.remaining == 4

    def test_create_wrap_bytestring(self):
        val = ByteBuffer(b'\x03\x05\x07\x09')
        assert val.position == 0
        assert val.capacity == 4
        assert val.remaining == 4
        assert val.get(4) == bytearray(b'\x03\x05\x07\x09')

    def test_create_wrap_bytearray(self):
        ba = bytearray(b'\x03\x05\x07\x09')
        val = ByteBuffer(ba)
        assert val.position == 0
        assert val.capacity == 4
        assert val.remaining == 4
        assert val.get(4) == ba

    def test_create_bad_type(self):
        self.assertRaises(TypeError, ByteBuffer, [])

    def test_get_int8(self):
        assert self.test_bb.getInt8() == 3
        assert self.test_bb.position == 1

    def test_get_int8_at_position(self):
        assert self.test_bb.getInt8(2) == 7
        assert self.test_bb.position == 0

    def test_get_int16(self):
        assert self.test_bb.getInt16() == 773
        assert self.test_bb.position == 2

    def test_get_int16_at_position(self):
        assert self.test_bb.getInt16(2) == 1801
        assert self.test_bb.position == 0

    def test_get_int32(self):
        assert self.test_bb.getInt32() == 50661129
        assert self.test_bb.position == 4

    def test_get_int32_at_position(self):
        assert self.test_bb.getInt32(2) == 118034707
        assert self.test_bb.position == 0

    def test_get_int64(self):
        assert self.test_bb.getInt64() == 217587892519900439
        assert self.test_bb.position == 8

    def test_get_int64_at_position(self):
        assert self.test_bb.getInt64(2) == 506955206711777569
        assert self.test_bb.position == 0

    def test_get(self):
        assert self.test_bb.get(4) == bytearray(b'\x03\x05\x07\x09')

    def test_get_at_position(self):
        assert self.test_bb.get(4, 2) == bytearray(b'\x07\x09\x11\x13')

    def test_put_int8(self):
        self.test_bb.putInt8(8)
        assert self.test_bb.get(10, 0) == bytearray(b'\x08\x05\x07\x09\x11\x13\x15\x17\x19\x21')

    def test_put_int8_at_position(self):
        self.test_bb.putInt8(8, 2)
        assert self.test_bb.get(10, 0) == bytearray(b'\x03\x05\x08\x09\x11\x13\x15\x17\x19\x21')

    def test_put_int16(self):
        self.test_bb.putInt16(2342)
        assert self.test_bb.get(10, 0) == bytearray(b'\t&\x07\x09\x11\x13\x15\x17\x19\x21')

    def test_put_int16_at_position(self):
        self.test_bb.putInt16(2342, 2)
        assert self.test_bb.get(10, 0) == bytearray(b'\x03\x05\t&\x11\x13\x15\x17\x19\x21')

    def test_put_int32(self):
        self.test_bb.putInt32(897324)
        assert self.test_bb.get(10, 0) == bytearray(b'\x00\r\xb1,\x11\x13\x15\x17\x19\x21')

    def test_put_int32_at_position(self):
        self.test_bb.putInt32(897324, 2)
        assert self.test_bb.get(10, 0) == bytearray(b'\x03\x05\x00\r\xb1,\x15\x17\x19\x21')

    def test_put_int64(self):
        self.test_bb.putInt64(7992478585871677467)
        assert self.test_bb.get(10, 0) == bytearray(b'n\xea\xfc\xed\x89\xcc8\x1b\x19\x21')

    def test_put_int64_at_position(self):
        self.test_bb.putInt64(7992478585871677467, 2)
        assert self.test_bb.get(10, 0) == bytearray(b'\x03\x05n\xea\xfc\xed\x89\xcc8\x1b')

    def test_put(self):
        self.test_bb.put(b'\x21\x23\x25\x27')
        assert self.test_bb.get(10, 0) == bytearray(b'\x21\x23\x25\x27\x11\x13\x15\x17\x19\x21')

    def test_put_at_position(self):
        self.test_bb.put(b'\x21\x23\x25\x27', 2)
        assert self.test_bb.get(10, 0) == bytearray(b'\x03\x05\x21\x23\x25\x27\x15\x17\x19\x21')

    def test_put_bad_type(self):
        self.assertRaises(TypeError, self.test_bb.put, 56)

    def test_underflow(self):
        self.assertRaises(EOFError, self.test_bb.getInt64, 3)

    def test_position_setter(self):
        self.test_bb.position = 3
        assert self.test_bb.position == 3
        assert self.test_bb.getInt8() == 9

    def test_position_setter_out_of_range(self):
        with self.assertRaises(IndexError):
            self.test_bb.position = 10

    def test_position_setter_bad_type(self):
        with self.assertRaises(TypeError):
            self.test_bb.position = 'nonint'

    def test_position_setter_out_of_range_negative(self):
        with self.assertRaises(IndexError):
            self.test_bb.position = -1

    def test_capacity(self):
        assert self.test_bb.capacity == 10

    def test_limit(self):
        assert self.test_bb.limit == 9

    def test_limit_setter(self):
        self.test_bb.limit = 4
        self.assertRaises(EOFError, self.test_bb.getInt64)

    def test_limit_setter_out_of_range(self):
        with self.assertRaises(IndexError):
            self.test_bb.limit = 10

    def test_limit_setter_bad_type(self):
        with self.assertRaises(TypeError):
            self.test_bb.limit = 'nonint'

    def test_rewind(self):
        self.test_bb.position = 4
        self.test_bb.rewind()
        assert self.test_bb.position == 0

    def test_duplicate(self):
        val = self.test_bb.duplicate()
        assert val._buffer is self.test_bb._buffer
        assert val._first == self.test_bb._first
        assert val.limit == self.test_bb.limit
        assert val.position == self.test_bb.position

    def test_slice(self):
        self.test_bb.position = 4
        val = self.test_bb.slice()
        assert val._buffer is self.test_bb._buffer
        assert val._first == 4
        assert val.limit == self.test_bb.limit
        assert val.position == 4

    def test_slice_rewind(self):
        self.test_bb.position = 4
        val = self.test_bb.slice()
        val.rewind()
        assert val.position == 4



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.list_offset_v2 import ListOffsetV2Request


class ListOffsetV1RequestTest(unittest.TestCase):
    def test_process_arguments(self):
        val = ListOffsetV2Request.process_arguments(['-1', 'false', 'topicname', '4,2', 'nexttopic', '9,3'])
        assert val == {'replica_id': -1,
                       'isolation_level': 0,
                       'topics': [{'topic': 'topicname', 'partitions': [{'partition': 4, 'timestamp': 2}]},
                                  {'topic': 'nexttopic', 'partitions': [{'partition': 9, 'timestamp': 3}]}]}

    def test_process_arguments_nonnumeric(self):
        self.assertRaises(ArgumentError, ListOffsetV2Request.process_arguments, ['foo', 'true', 'topicname', '4,2'])

    def test_process_arguments_nonbool(self):
        self.assertRaises(ArgumentError, ListOffsetV2Request.process_arguments, ['-1', 'notboolean', 'topicname', '4,2'])

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, ListOffsetV2Request.process_arguments, ['-1', 'true', 'topicname'])
